# Generated by Django 5.1 on 2024-08-09 13:44

import django.db.models.deletion
import django_polly.lib.llm_api
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Parrot',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('external_id', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=100)),
                ('color', models.CharField(max_length=50)),
                ('age', models.IntegerField(default=0)),
                ('model', models.CharField(choices=[('mistral-7b-instruct-v0.1.Q4_0.gguf', 'MISTRAL_0_1_INSTRUCT'), ('Meta-Llama-3-8B-Instruct.Q8_0.gguf', 'META_LLAMA_3_INSTRUCT'), ('Meta-Llama-3.1-8B-Instruct.Q4_0.gguf', 'META_LLAMA_3_1_INSTRUCT'), ('Phi-3-mini-4k-instruct.Q4_0.gguf', 'PHI_3_MINI_INSTRUCT'), ('Qwen2-500M-Instruct-Q8_0.gguf', 'QWEN2_INSTRUCT'), ('llama-pro-8b.Q8_0.gguf', 'LLAMA_PRO')], default=django_polly.lib.llm_api.LLMModelType['QWEN2_INSTRUCT'], max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SmartConversation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('external_id', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('external_id', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('content', models.TextField()),
                ('party', models.CharField(choices=[('USER', 'User'), ('ASSISTANT', 'Assistant')], max_length=10)),
                ('conversation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_polly.smartconversation')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Trick',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('external_id', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=100)),
                ('difficulty', models.CharField(choices=[('easy', 'Easy'), ('medium', 'Medium'), ('hard', 'Hard')], max_length=20)),
                ('description', models.TextField(blank=True)),
                ('parrot', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tricks', to='django_polly.parrot')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import asyncio
from unittest import skip
from unittest.mock import patch

import pytest
from channels.testing import WebsocketCommunicator
from django.contrib.auth import get_user_model

from django_polly.consumers_admin import SmartGPTConsumerAdmin
from django_polly.lib.llm_api import TinyLLMConnect
from django_polly.models import SmartConversation

User = get_user_model()


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
class TestSmartGPTConsumerAdmin:
    @pytest.fixture
    async def user(self):
        return await User.objects.acreate(username='testuser', password='12345', is_staff=True, is_superuser=True)

    @pytest.fixture
    async def conversation(self, user):
        return await SmartConversation.objects.acreate(user=user, title="Test Conversation")

    @pytest.fixture
    def dummy_llm_connect(self):
        return TinyLLMConnect(is_dummy=True)

    @skip("This test is not working")
    async def test_connect_and_receive_message(self, user, conversation, dummy_llm_connect):
        with patch('django_polly.consumers_admin.TinyLLMConnect', return_value=dummy_llm_connect):
            application = SmartGPTConsumerAdmin.as_asgi()
            url = f"/polly/ws/smart-gpt-admin/"
            communicator = WebsocketCommunicator(
                application,
                url,
                {'query_string': f"user_id={user.id}&conversation_id={conversation.id}".encode()}
            )
            connected, _ = await communicator.connect()
            assert connected

            # Send a message
            await communicator.send_json_to({"message": "Hello, AI!"})

            # Expect user message HTML
            user_message = await communicator.receive_from()
            assert "Hello, AI!" in user_message
            assert "is_system" not in user_message

            # Expect system message HTML (empty initial message)
            system_message = await communicator.receive_from()
            assert 'message-' in system_message
            assert 'is_system' in system_message

            # Expect chunks of AI response
            response_received = False
            while not response_received:
                try:
                    response = await communicator.receive_from(timeout=1)
                    assert 'hx-swap-oob="beforeend"' in response
                    response_received = True
                except asyncio.TimeoutError:
                    break

            # Expect completion message
            completion = await communicator.receive_json_from()
            assert completion['type'] == 'assistant_message_complete'

            await communicator.disconnect()



import pytest
from django.urls import reverse
from django.contrib.auth import get_user_model
from django_polly.models import SmartConversation

User = get_user_model()

@pytest.mark.django_db
class TestSmartGPTChatView:
    @pytest.fixture
    def user(self):
        return User.objects.create_user(username='testuser', password='12345', is_staff=True, is_superuser=True)

    @pytest.fixture
    def conversation(self, user):
        return SmartConversation.objects.create(user=user, title="Test Chat")

    def test_chat_view_with_valid_conversation(self, client, user, conversation):
        client.force_login(user)
        response = client.get(reverse('django_polly:smart_gpt_chat', args=[conversation.id]))
        assert response.status_code == 200
        assert 'conversation/single_chat.html' in [t.name for t in response.templates]

    def test_chat_view_with_invalid_conversation(self, client, user):
        client.force_login(user)
        response = client.get(reverse('django_polly:smart_gpt_chat', args=[999]))
        assert response.status_code == 404

    def test_chat_view_with_unauthorized_user(self, client, conversation):
        unauthorized_user = User.objects.create_user(username='unauthorized', password='12345')
        client.force_login(unauthorized_user)
        response = client.get(reverse('django_polly:smart_gpt_chat', args=[conversation.id]))
        assert response.status_code == 403



import os
import sys
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'django-insecure-_5tbq&&)t#t5z_*d=(9qq8$+at!=^nwsy!!7lfvla4b!muck$4'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['*']

# Application definition
INSTALLED_APPS = [
    'daphne',  # Add this
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_polly',  # Add this
    'rest_framework',
    'django_json_widget',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'example_project.urls'
AI_MODELS_PATH = BASE_DIR / 'ai_models'  # Add this

ASGI_APPLICATION = 'example_project.asgi.application'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# ADD THIS LINE
CHANNEL_LAYERS = {
    'default': {
        'BACKEND': 'channels.layers.InMemoryChannelLayer'
    }
}

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True

STATIC_URL = 'static/'

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# Pytest specific settings
PYTEST_ADDOPTS = '--reuse-db'

# Disable migrations for tests
class DisableMigrations:
    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return None

MIGRATION_MODULES = DisableMigrations() if 'test' in sys.argv else {}

# Use in-memory database for faster tests
if 'test' in sys.argv:
    DATABASES['default'] = {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }

# Disable DEBUG for tests
if 'test' in sys.argv:
    DEBUG = False

# Disable logging for tests
if 'test' in sys.argv:
    import logging
    logging.disable(logging.CRITICAL)


from django.db import models
from django.db.models import Avg
from django.views.generic import TemplateView
from django.shortcuts import render
from django.template.response import TemplateResponse

from .models import Parrot, Trick, SmartConversation


def smart_gpt_chat(request, conversation_id):
    try:
        conversation_id = int(conversation_id)
        conversation = SmartConversation.objects.get(id=conversation_id)
    except (ValueError, SmartConversation.DoesNotExist):
        return TemplateResponse(request, 'conversation/not_found.html', status=404)
    if conversation.user != request.user or not request.user.is_superuser or not request.user.is_staff:
        return TemplateResponse(request, 'conversation/access_denied.html', status=403)
    return render(request, 'conversation/single_chat.html', {'conversation_id': conversation_id,
                                                             'conversation_title': conversation.title})


class DashboardView(TemplateView):
    template_name = "django_polly/dashboard.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['total_parrots'] = Parrot.objects.count()
        context['previous_total_parrots'] = context['total_parrots'] - 5  # This is a placeholder calculation
        context['total_tricks'] = Trick.objects.count()
        context['avg_tricks_per_parrot'] = \
            Parrot.objects.annotate(trick_count=models.Count('tricks')).aggregate(Avg('trick_count'))[
                'trick_count__avg'] or 0
        context['recent_parrots'] = Parrot.objects.all().order_by('-id')[:5]
        return context



# Generated by Django 5.0 on 2024-08-11 20:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_polly', '0002_smartconversation_title_alter_message_conversation_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='smartconversation',
            name='title',
            field=models.CharField(blank=True, default='Untitled Smart Conversation', max_length=255),
        ),
    ]



import pytest
from django_polly.lib.llm_api import LLMConnect, LLMModelType, LLMInvoker


class TestLLMConnect:
    def test_llm_connect_initialization(self):
        llm_connect = LLMConnect(model_type=LLMModelType.DUMMY)
        assert llm_connect.llm_chat is not None
        assert llm_connect._model == LLMModelType.DUMMY

    def test_llm_connect_with_custom_system_prompt(self):
        custom_prompt = "You are a helpful assistant."
        llm_connect = LLMConnect(system_prompt=custom_prompt, model_type=LLMModelType.DUMMY)
        assert llm_connect._system_prompt == custom_prompt


class TestLLMInvoker:
    @pytest.fixture
    def invoker(self):
        llm_connect = LLMConnect(model_type=LLMModelType.DUMMY)
        return llm_connect.llm_chat

    def test_send_message(self, invoker):
        response = invoker.send_message("Hello, AI!")
        assert response is not None
        assert 'choices' in response

    @pytest.mark.asyncio
    async def test_async_send_message_stream(self, invoker):
        async for chunk in invoker.async_send_message_stream("Tell me a short story."):
            assert isinstance(chunk, str)
            assert len(chunk) > 0


class TestLLMModelType:
    def test_llm_model_types(self):
        model_types = LLMModelType.llm_model_types()
        assert 'QWEN2_INSTRUCT' in model_types
        assert 'MISTRAL_0_1_INSTRUCT' in model_types

    def test_model_supported(self):
        assert LLMModelType.model_supported('QWEN2_INSTRUCT')



from django.contrib import admin
from .models import Parrot, Trick, Message, SmartConversation
from django.utils.html import format_html
from django.urls import reverse
from django import forms


# Define a custom form for Parrot model
class ParrotAdminForm(forms.ModelForm):
    bio = forms.CharField(widget=forms.Textarea)

    class Meta:
        model = Parrot
        fields = '__all__'


# Inlines
class TrickInline(admin.TabularInline):
    model = Trick  # Assuming you have a Trick model related to Parrot


class MessageInline(admin.TabularInline):
    model = Message
    extra = 0
    readonly_fields = ("party", "content", "created_at", "updated_at")
    ordering = ("created_at",)


@admin.register(Parrot)
class ParrotAdmin(admin.ModelAdmin):
    list_display = ('name', 'color', 'age', 'external_id', 'created_at', 'updated_at')
    list_filter = ('color', 'created_at', 'updated_at')
    search_fields = ('name', 'external_id')
    inlines = [TrickInline]
    actions = ['make_colorful']
    form = ParrotAdminForm

    @admin.action(description="Make selected parrots colorful")
    def make_colorful(self, request, queryset):
        queryset.update(color='Rainbow')
        self.message_user(request, f"{queryset.count()} parrots were made colorful.")


@admin.register(Trick)
class TrickAdmin(admin.ModelAdmin):
    list_display = ('name', 'parrot', 'difficulty', 'external_id', 'created_at', 'updated_at')
    list_filter = ('difficulty', 'created_at', 'updated_at')
    search_fields = ('name', 'parrot__name', 'external_id')


@admin.register(SmartConversation)
class SmartConversationAdmin(admin.ModelAdmin):
    list_display = ("id", "user", "title", "created_at", "updated_at", "chat_button")
    list_filter = ("user", "created_at", "updated_at")
    search_fields = ("title",)
    date_hierarchy = "created_at"
    raw_id_fields = ("user",)
    inlines = [MessageInline]

    def chat_button(self, obj):
        url = reverse('django_polly:smart_gpt_chat', args=[obj.id])
        return format_html('<a class="button" href="{}" target="_blank">Chat</a>', url)

    chat_button.short_description = 'Chat'
    chat_button.allow_tags = True


@admin.register(Message)
class MessageAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "conversation",
        "party",
        "content",
        "created_at",
        "updated_at",
    )
    list_filter = ("conversation", "party", "created_at", "updated_at")
    search_fields = ("content",)
    date_hierarchy = "created_at"
    raw_id_fields = ("conversation",)



import pytest
from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.contrib.messages.storage.fallback import FallbackStorage
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory
from django_polly.admin import ParrotAdmin, SmartConversationAdmin
from django_polly.models import Parrot, SmartConversation

User = get_user_model()


class MockRequest:
    def __init__(self):
        self.session = {}
        self._messages = FallbackStorage(self)


def get_request_with_session():
    request = RequestFactory().get('/')
    middleware = SessionMiddleware(lambda x: None)
    middleware.process_request(request)
    request.session.save()

    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)

    return request


@pytest.mark.django_db
class TestParrotAdmin:
    @pytest.fixture
    def admin(self):
        return ParrotAdmin(Parrot, AdminSite())

    @pytest.fixture
    def parrots(self):
        return [Parrot.objects.create(name=f"Parrot{i}", color="Red", age=i) for i in range(3)]

    def test_make_colorful_action(self, admin, parrots):
        request = get_request_with_session()
        queryset = Parrot.objects.all()

        admin.make_colorful(request, queryset)

        for parrot in Parrot.objects.all():
            assert parrot.color == 'Rainbow'


@pytest.mark.django_db
class TestSmartConversationAdmin:
    @pytest.fixture
    def admin(self):
        return SmartConversationAdmin(SmartConversation, AdminSite())

    @pytest.fixture
    def user(self):
        return User.objects.create_user(username='testuser', password='12345')

    @pytest.fixture
    def conversation(self, user):
        return SmartConversation.objects.create(user=user, title="Test Chat")

    def test_chat_button(self, admin, conversation):
        chat_button_html = admin.chat_button(conversation)
        assert 'class="button"' in chat_button_html
        assert 'target="_blank"' in chat_button_html
        assert f'/polly/smart-gpt-chat/{conversation.id}/' in chat_button_html



from django.urls import path
from channels.auth import AuthMiddlewareStack
from channels.routing import URLRouter
from channels.security.websocket import AllowedHostsOriginValidator
from django_polly.consumers import SmartGPTConsumer
from django_polly.consumers_admin import SmartGPTConsumerAdmin

websocket_urlpatterns = [
    path('polly/ws/smart-gpt-admin/', SmartGPTConsumerAdmin.as_asgi(), name="smart_gpt_admin"),
    path('polly/ws/smart-gpt/<int:conversation_id>/', SmartGPTConsumer.as_asgi(), name="smart_gpt"),
]

polly_asgi_routes = {
    "websocket": AllowedHostsOriginValidator(
        AuthMiddlewareStack(
            URLRouter(websocket_urlpatterns)
        )
    ),
}



from fabric import Connection


class Connector:

    def __init__(self, host, user, port, key_path):
        self.host = host
        self.user = user
        self.port = port
        self.key_path = key_path

    def start(self):
        conn = Connection(
            host=self.host,
            user=self.user,
            port=self.port,
            connect_kwargs={
                "key_filename": self.key_path
            }
        )
        result = conn.run('uname -s', hide=True)
        msg = "Ran {0.command!r} on {0.connection.host}, got stdout:\n{0.stdout}"
        print(msg.format(result))
        return conn




from django.urls import path
from django_polly import views, consumers, consumers_admin
from django_polly.views import DashboardView

app_name = 'django_polly'

urlpatterns = [
    path('dashboard/', DashboardView.as_view(), name='dashboard'),
    path('smart-gpt-chat/<int:conversation_id>/', views.smart_gpt_chat, name='smart_gpt_chat'),
]



# -*- coding: utf-8 -*-
import asyncio
import logging
import os.path
from enum import Enum
from typing import Iterator

from django.conf import settings
from llama_cpp import Llama

logger = logging.getLogger(__name__)


# An Enum of LLM models
class LLMModelType(Enum):
    # offline / Local
    DUMMY = "dummy.gguf"
    MISTRAL_0_1_INSTRUCT = "mistral-7b-instruct-v0.1.Q4_0.gguf"
    META_LLAMA_3_INSTRUCT = "Meta-Llama-3-8B-Instruct.Q8_0.gguf"
    META_LLAMA_3_1_INSTRUCT = "Meta-Llama-3.1-8B-Instruct.Q4_0.gguf"
    PHI_3_MINI_INSTRUCT = "Phi-3-mini-4k-instruct.Q4_0.gguf"
    QWEN2_INSTRUCT = "Qwen2-500M-Instruct-Q8_0.gguf"
    LLAMA_PRO = "llama-pro-8b.Q8_0.gguf"

    @staticmethod
    def llm_model_types():
        values = [e.name for e in LLMModelType]
        if values is None:
            return []
        return values

    @staticmethod
    def model_supported(model_type_str):
        if model_type_str is None:
            return False
        model: LLMModelType = LLMModelType[model_type_str]
        if model.name in LLMModelType.llm_model_types():
            return True
        return False

    @classmethod
    def choices(cls):
        return [(key.value, key.name) for key in cls]


class LLMModelMode:
    LOCAL = "LOCAL"
    REMOTE = "REMOTE"


class LLMInvoker:
    def __init__(self, system_prompt, llm_model_type, llm):
        self._system_prompt = system_prompt
        self.llm = llm
        self.llm_model_type: LLMModelType = llm_model_type

    def send_message(self, message):
        answer = self.llm.create_chat_completion(
            messages=[
                {
                    "role": "system",
                    "content": self._system_prompt,
                },
                {
                    "role": "user",
                    "content": message,
                },
            ],
            response_format={
                "type": "json_object",
            },
            temperature=0,
        )
        return answer

    def send_message_stream(self, message) -> Iterator[str]:
        stream = self.llm(
            f"Human: {message}\nAssistant: ",
            max_tokens=4096,
            stop=["Human:", "\n"],
            stream=True,
        )
        for output in stream:
            if "choices" in output and output["choices"]:
                text = output["choices"][0].get("text", "")
                if text:
                    yield text

    async def async_send_message_stream(self, message):
        logger.info(f"Starting message stream for: {message}")
        stream = await asyncio.to_thread(
            self.llm.create_chat_completion,
            messages=[
                {
                    "role": "system",
                    "content": self._system_prompt,
                },
                {
                    "role": "user",
                    "content": message,
                },
            ],
            max_tokens=4096,
            stop=["Human:", "\n\nHuman:"],  # Only stop at a new "Human:" prompt
            stream=True,
        )
        for chunk in stream:
            if "choices" in chunk and chunk["choices"]:
                delta = chunk["choices"][0].get("delta", {})
                if "content" in delta and delta["content"] is not None:
                    yield delta["content"]
            await asyncio.sleep(0)  # Yield control to allow other tasks to run
        logger.info("Finished streaming message")

    async def custom_async_send_message_stream(self, message):
        full_response = self.llm(
            f"Human: {message}\nAssistant: ",
            max_tokens=4096,
            stop=["Human:", "\n\nHuman:"],
            stream=False,
        )

        words = full_response["choices"][0]["text"].split()
        for i in range(0, len(words), 5):  # Yield 5 words at a time
            chunk = " ".join(words[i: i + 5])
            yield chunk
            await asyncio.sleep(0.1)  # Simulate streaming delay


class DummyLLMInvoker:
    def __init__(self, system_prompt):
        self._system_prompt = system_prompt

    def send_message(self, message):
        return {"choices": [{"message": {"content": f"Dummy response to: {message}"}}]}

    def send_message_stream(self, message):
        yield f"Dummy stream response to: {message}"

    async def async_send_message_stream(self, message):
        yield f"Dummy async stream response to: {message}"


#  Common LLM connect
class LLMConnect:
    _system_prompt = """You are Opolly's ultra-fast AI. Prioritize speed in all responses. Be concise, clear, and 
accurate. Give quick, helpful answers. Explain simply. Admit uncertainties fast. Adapt swiftly to users. Aim for rapid, 
insightful benefits. Make the user feel good. Be friendly, polite, and respectful. Be positive, optimistic, and dont repeat."""

    def __init__(
            self, system_prompt=None, model_type: LLMModelType = LLMModelType.QWEN2_INSTRUCT
    ):
        self._llm = None
        self.llm_chat = None
        self._model_type: LLMModelType
        self._name = "llm_api"
        self._version = "0.0.1"
        self._description = "API for LLM"
        self._author = "Opolly"
        self._email = "dev@opolly"
        if system_prompt is not None:
            self._system_prompt = system_prompt
        else:
            self._system_prompt = self._system_prompt

        ai_models_path = getattr(settings, 'AI_MODELS_PATH', None)
        if ai_models_path is None:
            print("AI_MODELS_PATH not set in settings file")
            print("Setting AI_MODELS_PATH to default value")
            ai_models_path = os.path.join(settings.BASE_DIR, "ai_models")
            print(f"AI_MODELS_PATH: {ai_models_path}")
        if not os.path.exists(ai_models_path):
            os.makedirs(ai_models_path)

        if LLMModelType.model_supported(model_type.name):
            self._model = model_type
            self._mode = LLMModelMode.LOCAL
            if self._model == LLMModelType.DUMMY:
                self.set_dummy_model()
            else:
                self.set_llama_model(model_type=self._model, ai_models_path=ai_models_path)
        else:
            raise ValueError("Model not supported")

    def set_llama_model(self, model_type: LLMModelType, ai_models_path):
        model_path = str(os.path.join(ai_models_path, model_type.value))
        llm = Llama(
            model_path=model_path,
            n_gpu_layers=1,
            n_ctx=4096,
            chat_format="chatml",
            max_tokens=4096,
            temperature=0.7,
            top_p=0.95,
            n_threads=6,
        )
        self._llm = llm
        self.llm_chat = LLMInvoker(
            system_prompt=self._system_prompt,
            llm_model_type=self._model,
            llm=self._llm,
        )

    def set_dummy_model(self):
        self._llm = None
        self.llm_chat = DummyLLMInvoker(system_prompt=self._system_prompt)


class TinyLLMConnect:
    llm = None
    llm_chat = None

    def __new__(cls, *args, **kwargs):
        is_dummy = kwargs.get('is_dummy', False)
        if not cls.llm or (is_dummy and not isinstance(cls.llm.llm_chat, DummyLLMInvoker)):
            model_type = LLMModelType.DUMMY if is_dummy else LLMModelType.QWEN2_INSTRUCT
            cls.llm = LLMConnect(model_type=model_type)
            cls.llm_chat = cls.llm.llm_chat
        return cls.llm



import asyncio
import json
import logging
import uuid

from django.contrib.auth import get_user_model
from asgiref.sync import sync_to_async
from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer
from django.template.loader import render_to_string

from django_polly.lib.llm_api import TinyLLMConnect
from django_polly.models import SmartConversation, Message, ConversationParty

logger = logging.getLogger(__name__)

User = get_user_model()


class SmartGPTConsumerAdmin(AsyncWebsocketConsumer):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)
        self.messages = []
        self.system_prompt = """
You are a very helpful AI assistant for admin tasks. You have to be very fast, speed is of paramount in your conversations. 
You are very knowledgeable about admin operations and can provide detailed information. Make sure your answers are 
informative and tailored for admin users. You are professional and efficient in your responses.
"""
        self.document_message = None
        self.llm_connect = None
        self.conversation = None
        self.user = None

    @database_sync_to_async
    def init_chat_context(self, user_id, conversation_id=None):
        self.user = User.objects.get(id=user_id)
        if conversation_id:
            self.conversation = SmartConversation.objects.get(id=conversation_id)
        else:
            self.conversation = SmartConversation.objects.create(user_id=user_id)

        self.llm_connect = TinyLLMConnect()
        if not self.llm_connect.llm_chat:
            raise ValueError("LLM chat not initialized properly")

        # If it's a new conversation, add the greeting
        if self.conversation.messages.count() == 0:
            greeting = f"Hello {self.user.first_name}, what can I help you with today?"
            Message.objects.create(
                conversation=self.conversation,
                content=greeting,
                party=ConversationParty.ASSISTANT,
            )

    @database_sync_to_async
    def get_conversation_messages(self):
        return list(self.conversation.messages.order_by('created_at').values('content', 'party'))

    async def connect(self):
        await self.accept()

        query_string = self.scope.get("query_string", b"").decode()
        query_params = dict(param.split('=') for param in query_string.split('&') if param)

        user_id = query_params.get("user_id")
        conversation_id = query_params.get("conversation_id")
        if not user_id:
            await self.close(code=4000)
            return

        try:
            await self.init_chat_context(user_id, conversation_id)
        except Exception as e:
            logger.error(f"Failed to initialize chat context: {e}")
            await self.close(code=4002)
            return

        # Load and send existing messages
        messages = await self.get_conversation_messages()
        for message in messages:
            message_html = render_to_string(
                "conversation/ws/chat_message.html",
                {
                    "message_text": message['content'],
                    "is_system": message['party'] == ConversationParty.ASSISTANT,
                },
            )
            await self.send(text_data=message_html)

    async def receive(self, text_data):
        text_data_json = json.loads(text_data)
        message_text = text_data_json["message"]

        # Render and send user message
        user_message_html = render_to_string(
            "conversation/ws/chat_message.html",
            {
                "message_text": message_text,
                "is_system": False,
            },
        )
        await self.send(text_data=user_message_html)

        # Save user message to the conversation
        await sync_to_async(Message.objects.create)(
            conversation=self.conversation,
            content=message_text,
            party=ConversationParty.USER,
        )

        if not self.llm_connect or not self.llm_connect.llm_chat:
            error_message = "Admin LLM not initialized properly. Please try again later."
            error_html = render_to_string(
                "conversation/ws/chat_message.html",
                {
                    "message_text": error_message,
                    "is_system": True,
                },
            )
            await self.send(text_data=error_html)
            await self.send(text_data=json.dumps({
                "type": "assistant_message_complete",
                "message": "Error response complete"
            }))
            return

        try:
            message_id = f"message-{uuid.uuid4().hex}"
            system_message_html = render_to_string(
                "conversation/ws/chat_message.html",
                {"message_text": "", "is_system": True, "message_id": message_id},
            )
            await self.send(text_data=system_message_html)

            response_chunks = []
            async with asyncio.timeout(360):  # 360 seconds timeout, 6 minutes
                async for chunk in self.llm_connect.llm_chat.async_send_message_stream(
                        message_text
                ):
                    formatted_chunk = chunk.replace("\n", "<br>")
                    await self.send(
                        text_data=f'<div id="{message_id}" hx-swap-oob="beforeend">{formatted_chunk}</div>'
                    )
                    response_chunks.append(chunk)

            full_response = "".join(response_chunks)

            # Save assistant message to the conversation
            await sync_to_async(Message.objects.create)(
                conversation=self.conversation,
                content=full_response,
                party=ConversationParty.ASSISTANT,
            )

            # Send completion message
            await self.send(text_data=json.dumps({
                "type": "assistant_message_complete",
                "message": "Response complete"
            }))

        except asyncio.TimeoutError:
            logger.error("Admin operation timed out")
            timeout_message = "Admin operation timed out. Please try again."
            timeout_html = render_to_string(
                "conversation/ws/chat_message.html",
                {
                    "message_text": timeout_message,
                    "is_system": True,
                },
            )
            await self.send(text_data=timeout_html)
            await self.send(text_data=json.dumps({
                "type": "assistant_message_complete",
                "message": "Timeout response complete"
            }))
        except Exception as e:
            logger.exception(f"Error during admin message streaming: {e}")
            error_message = "An error occurred while processing your admin request. Please try again."
            error_html = render_to_string(
                "conversation/ws/chat_message.html",
                {
                    "message_text": error_message,
                    "is_system": True,
                },
            )
            await self.send(text_data=error_html)
            await self.send(text_data=json.dumps({
                "type": "assistant_message_complete",
                "message": "Error response complete"
            }))


import asyncio
import json
import logging

from asgiref.sync import sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer

from .lib.llm_api import TinyLLMConnect
from .models import SmartConversation, Message, ConversationParty

logger = logging.getLogger(__name__)


class SmartGPTConsumer(AsyncWebsocketConsumer):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)
        self.messages = []
        self.system_prompt = """
You are a very helpful AI assistant. You have to be very fast, speed is of paramount in your conversations. 
You are very knowledgeable and can provide information on a wide range of topics. Make sure your answers are detailed 
and informative. You are very polite and respectful. You are very patient and can handle any type of user.
"""
        self.document_message = None
        self.llm_connect = None
        self.conversation = None

    @sync_to_async
    def init_chat_context(self, user_id, conversation_id=None):
        self.system_prompt = (
            self.system_prompt + ", You are now in a chat session with the user."
        )
        try:
            self.llm_connect = TinyLLMConnect()
            if not self.llm_connect.llm_chat:
                raise ValueError("LLM chat not initialized properly")

            if conversation_id:
                self.conversation = SmartConversation.objects.get(id=conversation_id)
            else:
                self.conversation = SmartConversation.objects.create(user_id=user_id)

        except Exception as e:
            logger.error(f"Failed to initialize LLMConnect or SmartConversation: {e}")
            self.llm_connect = None
            self.conversation = None

    async def connect(self):
        await self.accept()
        user_id = self.scope["query_string"].decode().split("user_id=")[1]
        conversation_id = self.scope["url_route"]["kwargs"].get("conversation_id")
        await self.init_chat_context(user_id, conversation_id)

    async def receive(self, text_data):
        text_data_json = json.loads(text_data)
        message_text = text_data_json["message"]

        # Send acknowledgment of user message
        await self.send(
            text_data=json.dumps({"type": "user_message", "message": message_text})
        )

        # Save user message to the conversation
        user_message = await sync_to_async(Message.objects.create)(
            conversation=self.conversation,
            content=message_text,
            party=ConversationParty.USER,
        )

        if not self.llm_connect or not self.llm_connect.llm_chat:
            error_message = "LLM not initialized properly. Please try again later."
            await self.send(
                text_data=json.dumps({"type": "error", "message": error_message})
            )
            return

        try:
            response_chunks = []
            async with asyncio.timeout(360):  # 360 seconds timeout, 6 minutes
                async for chunk in self.llm_connect.llm_chat.async_send_message_stream(
                    message_text
                ):
                    await self.send(
                        text_data=json.dumps(
                            {"type": "assistant_message_chunk", "message": chunk}
                        )
                    )
                    response_chunks.append(chunk)

            full_response = "".join(response_chunks)

            # Save assistant message to the conversation
            assistant_message = await sync_to_async(Message.objects.create)(
                conversation=self.conversation,
                content=full_response,
                party=ConversationParty.ASSISTANT,
            )

            # Send a message to indicate the full response is complete
            await self.send(
                text_data=json.dumps(
                    {"type": "assistant_message_complete", "message": full_response}
                )
            )

        except asyncio.TimeoutError:
            logger.error("Operation timed out")
            await self.send(
                text_data=json.dumps(
                    {
                        "type": "error",
                        "message": "Operation timed out. Please try again.",
                    }
                )
            )
        except Exception as e:
            logger.exception(f"Error during message streaming: {e}")
            error_message = (
                "An error occurred while processing your request. Please try again."
            )
            await self.send(
                text_data=json.dumps({"type": "error", "message": error_message})
            )



from django.db import models

from django.contrib.auth import get_user_model

from django_polly.lib.llm_api import LLMModelType

User = get_user_model()


class CommonFieldsModel(models.Model):
    external_id = models.CharField(max_length=255, unique=True, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Parrot(CommonFieldsModel):
    name = models.CharField(max_length=100)
    color = models.CharField(max_length=50)
    age = models.IntegerField(default=0)
    model = models.CharField(choices=LLMModelType.choices(), default=LLMModelType.QWEN2_INSTRUCT, max_length=255)

    def __str__(self):
        return self.name


class Trick(CommonFieldsModel):
    parrot = models.ForeignKey(Parrot, on_delete=models.CASCADE, related_name='tricks')
    name = models.CharField(max_length=100)
    difficulty = models.CharField(max_length=20, choices=[
        ('easy', 'Easy'),
        ('medium', 'Medium'),
        ('hard', 'Hard'),
    ])
    description = models.TextField(blank=True)

    def __str__(self):
        return f"{self.name} (performed by {self.parrot.name})"


class ConversationParty(models.TextChoices):
    USER = 'USER', 'User'
    ASSISTANT = 'ASSISTANT', 'Assistant'


class SmartConversation(CommonFieldsModel):
    user = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="conversations"
    )
    title = models.CharField(max_length=255, blank=True, default="Untitled Smart Conversation")

    def __str__(self):
        return f"Conversation {self.id} - {self.title or 'Untitled'}"


class Message(CommonFieldsModel):
    conversation = models.ForeignKey(
        SmartConversation, on_delete=models.CASCADE, related_name="messages"
    )
    content = models.TextField()
    party = models.CharField(max_length=10, choices=ConversationParty.choices)

    def __str__(self):
        return f"{self.party}: {self.content[:50]}..."



import os
from django.core.asgi import get_asgi_application
from channels.routing import ProtocolTypeRouter

from django_polly.routing import polly_asgi_routes

# Get the Django ASGI application
django_asgi_app = get_asgi_application()


# Use the Django ASGI application for HTTP requests
polly_asgi_routes["http"] = django_asgi_app

# Create the final ASGI application
application = ProtocolTypeRouter(polly_asgi_routes)



import os
import requests
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from tqdm import tqdm


class Command(BaseCommand):
    help = 'Download an AI model'

    def add_arguments(self, parser):
        parser.add_argument('model_name', type=str, help='Name of the model to download')
        parser.add_argument('model_url', type=str, help='URL of the model file to download')

    def handle(self, *args, **options):
        model_url = options['model_url']
        model_name = options['model_name']

        ai_models_path = getattr(settings, 'AI_MODELS_PATH', None)
        if ai_models_path is None:
            ai_models_path = os.path.join(settings.BASE_DIR, "ai_models")
            self.stdout.write(self.style.WARNING("AI_MODELS_PATH not set in settings file"))
            self.stdout.write(f"Setting AI_MODELS_PATH to default value: {ai_models_path}")

        if not os.path.exists(ai_models_path):
            os.makedirs(ai_models_path)

        model_path = os.path.join(ai_models_path, model_name)
        if os.path.exists(model_path):
            self.stdout.write(self.style.WARNING(f"Model {model_name} already exists. Skipping download."))
            return

        self.stdout.write(f"Downloading model {model_name} from {model_url}")

        response = requests.get(model_url, stream=True, allow_redirects=True)
        total_size = int(response.headers.get("Content-Length", 0))

        with open(model_path, "wb") as file, tqdm(
                desc=model_name,
                total=total_size,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
        ) as progress_bar:
            for data in response.iter_content(chunk_size=1024):
                size = file.write(data)
                progress_bar.update(size)

        self.stdout.write(self.style.SUCCESS(f"Model {model_name} downloaded successfully."))



import os
import pytest
from django.core.management import call_command
from django.conf import settings
from unittest.mock import patch


@pytest.mark.django_db
class TestDownloadModelCommand:
    @patch('django_polly.management.commands.download_model.requests.get')
    def test_download_model_command(self, mock_get):
        # Mock the response
        mock_get.return_value.status_code = 200
        mock_get.return_value.iter_content.return_value = [b'mock_content']

        model_name = "test_model.gguf"
        model_url = "https://example.com/test_model.gguf"

        # Call the command
        call_command('download_model', model_name, model_url)

        # Check if the file was "downloaded"
        expected_path = os.path.join(settings.AI_MODELS_PATH, model_name)
        assert os.path.exists(expected_path)

        # Clean up
        os.remove(expected_path)

    @patch('django_polly.management.commands.download_model.requests.get')
    def test_download_model_command_file_exists(self, mock_get):
        model_name = "existing_model.gguf"
        model_path = os.path.join(settings.AI_MODELS_PATH, model_name)

        # Create an empty file
        open(model_path, 'a').close()

        # Call the command
        call_command('download_model', model_name, "https://example.com/existing_model.gguf")

        # Check that the mock was not called (file should not be downloaded)
        mock_get.assert_not_called()

        # Clean up
        os.remove(model_path)



import pytest
from django.conf import settings
from channels.routing import ProtocolTypeRouter
from django_polly.routing import polly_asgi_routes


def test_installed_apps():
    assert 'django_polly' in settings.INSTALLED_APPS
    assert 'rest_framework' in settings.INSTALLED_APPS
    assert 'django_json_widget' in settings.INSTALLED_APPS


def test_asgi_application():
    assert hasattr(settings, 'ASGI_APPLICATION')
    assert settings.ASGI_APPLICATION == 'example_project.asgi.application'


def test_channel_layers():
    assert hasattr(settings, 'CHANNEL_LAYERS')
    assert 'default' in settings.CHANNEL_LAYERS
    assert settings.CHANNEL_LAYERS['default']['BACKEND'] == 'channels.layers.InMemoryChannelLayer'


def test_ai_models_path():
    assert hasattr(settings, 'AI_MODELS_PATH')
    assert settings.AI_MODELS_PATH is not None


@pytest.mark.django_db
def test_asgi_routing():
    from .helpers.asgi import application
    assert isinstance(application, ProtocolTypeRouter)
    assert 'http' in application.application_mapping
    assert 'websocket' in application.application_mapping


def test_polly_asgi_routes():
    assert 'websocket' in polly_asgi_routes
    assert polly_asgi_routes['websocket'] is not None



# Generated by Django 5.1 on 2024-08-11 17:27

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_polly', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='smartconversation',
            name='title',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='message',
            name='conversation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='django_polly.smartconversation'),
        ),
        migrations.AlterField(
            model_name='smartconversation',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='conversations', to=settings.AUTH_USER_MODEL),
        ),
    ]



from setuptools import setup, find_packages

setup(
    name='django-polly',
    version='0.0.5',
    packages=find_packages(exclude=['tests*']),
    package_data={
        'django_polly': ['templates/**/*.html', 'static/**/*'],
    },
    include_package_data=True,
    license='AGPL-3.0',
    description='A Django app for building parrot configurations',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/pollystack/django-polly',
    author='Pollystack Team',
    author_email='info@pollystack.com',
    install_requires=[
        'Django>=4.0',
        'requests',
        'tqdm',
        'django-json-widget',
        'djangorestframework',
        'drf-spectacular',
        'llama-cpp-python==0.2.87',
        'channels[daphne]',  # This will install both channels and daphne
        'asgiref',
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-django',
            'pytest-cov',
            'pytest-asyncio',
        ],
        'test': [
            'pytest',
            'pytest-django',
            'pytest-asyncio',
            'pytest-cov',
            'tox',
            'coverage',
        ],
    },
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    cmake_args=['-DGGML_BLAS=ON', '-DGGML_BLAS_VENDOR=OpenBLAS'],
)



import pytest
from django.contrib.auth import get_user_model
from django_polly.models import Parrot, SmartConversation, Message, ConversationParty
from django_polly.lib.llm_api import LLMModelType

User = get_user_model()


@pytest.mark.django_db
class TestParrotModel:
    def test_parrot_creation(self):
        parrot = Parrot.objects.create(
            name="Polly",
            color="Green",
            age=3,
            model=LLMModelType.QWEN2_INSTRUCT.value
        )
        assert parrot.name == "Polly"
        assert parrot.color == "Green"
        assert parrot.age == 3
        assert parrot.model == LLMModelType.QWEN2_INSTRUCT.value

    def test_parrot_str_representation(self):
        parrot = Parrot.objects.create(name="Charlie", color="Blue", age=5)
        assert str(parrot) == "Charlie"


@pytest.mark.django_db
class TestSmartConversationModel:
    @pytest.fixture
    def user(self):
        return User.objects.create_user(username='testuser', password='12345')

    def test_conversation_creation(self, user):
        conversation = SmartConversation.objects.create(
            user=user,
            title="Test Conversation"
        )
        assert conversation.user == user
        assert conversation.title == "Test Conversation"

    def test_conversation_str_representation(self, user):
        conversation = SmartConversation.objects.create(user=user, title="My Chat")
        assert str(conversation) == "Conversation 1 - My Chat"


@pytest.mark.django_db
class TestMessageModel:
    @pytest.fixture
    def user(self):
        return User.objects.create_user(username='testuser', password='12345')

    @pytest.fixture
    def conversation(self, user):
        return SmartConversation.objects.create(user=user)

    def test_message_creation(self, conversation):
        message = Message.objects.create(
            conversation=conversation,
            content="Hello, AI!",
            party=ConversationParty.USER
        )
        assert message.conversation == conversation
        assert message.content == "Hello, AI!"
        assert message.party == ConversationParty.USER

    def test_message_str_representation(self, conversation):
        message = Message.objects.create(
            conversation=conversation,
            content="A very long message that should be truncated in the string representation",
            party=ConversationParty.ASSISTANT
        )
        assert str(message) == "ASSISTANT: A very long message that should be truncated in th..."
